import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures


def BAG(train_pred, test_pred, train_y, test_y):
    """
    Brain Age Gap Correction

    Parameters:
    train_pred (numpy array): Predicted age on the training set.
    test_pred (numpy array): Predicted age on the test set.
    train_y (numpy array): Actual age in the training set.
    test_y (numpy array): Actual age in the test set.

    Returns:
    corrected_age (numpy array): Corrected predicted age on the test set.
    train_delta_age (numpy array): Difference between predicted and actual age on the training set.
    line_reg (LinearRegression object): Trained regression model.
    """

    line_reg = LinearRegression()
    poly = PolynomialFeatures(degree=1, include_bias=True)
    line_reg_features = poly.fit_transform(train_y.reshape(-1, 1))
    line_reg.fit(line_reg_features, train_pred.reshape(-1, 1))

    test_features = poly.transform(test_y.reshape(-1, 1))
    corrected_delta_age = (
        test_pred - np.array(line_reg.predict(test_features)).flatten()
    )
    corrected_age = test_y + corrected_delta_age

    return corrected_age, corrected_delta_age



from neuroimager.utils.atlas import (
    split_schafer_names,
    combine_atlases,
    combine_probabilistic_atlases,
    filter_rois,
)
from neuroimager.utils.fc import (
    flatten_lower_triangular,
    unflatten_lower_triangular,
    filter_fcs,
    extract_bipartite_matrix,
    average_nodes,
)

__all__ = [
    "split_schafer_names",
    "combine_atlases",
    "combine_probabilistic_atlases",
    "filter_rois",
    "flatten_lower_triangular",
    "unflatten_lower_triangular",
    "filter_fcs",
    "extract_bipartite_matrix",
    "average_nodes",
]



import numpy as np
import pandas as pd


def convert_hmm(hmm):
    """
    ***Only tested for FMRI DATA***
    Convert the `hmm` object derived from the following command of HMM-MAR Toolbox
    `[hmm, Gamma, Xi, vpath] = hmmmar(f,T,options);`
    to a Python Dictionary for further processing
    """
    hmm_dict = {}
    hmm_param = ["train", "K", "prior", "Dir_alpha", "Pi", "Dir2d_alpha", "P", "state"]
    train_param = [
        "order",
        "verbose",
        "cyc",
        "initrep",
        "initcyc",
        "standardise",
        "inittype",
        "zeromean",
        "covtype",
        "useParallel",
        "repetitions",
        "Fs",
        "K",
        "onpower",
        "leida",
        "embeddedlags",
        "pca",
        "pca_spatial",
        "lowrank",
        "varimax",
        "maxFOth",
        "pcamar",
        "pcapred",
        "filter",
        "detrend",
        "downsample",
        "leakagecorr",
        "sequential",
        "standardise_pc",
        "regularisation",
        "Gamma_constraint",
        "acrosstrial_constrained",
        "plotGamma",
        "id_mixture",
        "episodic",
        "cluster",
        "initcriterion",
        "meancycstop",
        "cycstogoafterevent",
        "initTestSmallerK",
        "stopcriterion",
        "tol",
        "behaviour",
        "tudamonitoring",
        "tuda",
        "distribution",
        "grouping",
        "Pstructure",
        "Pistructure",
        "dropstates",
        "priorcov_rate",
        "timelag",
        "exptimelag",
        "orderoffset",
        "symmetricprior",
        "uniqueAR",
        "S",
        "prior",
        "Sind",
        "orders",
        "maxorder",
        "DirichletDiag",
        "PriorWeightingP",
        "PriorWeightingPi",
        "hmm",
        "fehist",
        "updateObs",
        "updateGamma",
        "updateP",
        "decodeGamma",
        "keepS_W",
        "useMEX",
        "ndim",
        "active",
    ]
    prior_param = ["Dir2d_alpha", "Dir_alpha"]
    omega_param = ["Gam_shape", "Gam_rate"]
    sub_param = {"train": train_param, "prior": prior_param, "Omega": omega_param}

    hmm = hmm["hmm"][0][0]
    for i, param in enumerate(hmm_param):
        data = hmm[i]

        if param in sub_param.keys():
            data = data[0][0]
            tem_dict = {}

            for j, sub in enumerate(sub_param[param]):
                tem_dict[sub] = data[j]
            hmm_dict[param] = tem_dict
            if param == "train":
                if tem_dict["covtype"] == "uniquefull":
                    hmm_param.append("Omega")
                    prior_param.append("Omega")
                    omega_param.append("Gam_irate")
                    unique = True
                elif tem_dict["covtype"] == "uniquediag":
                    hmm_param.append("Omega")
                    prior_param.append("Omega")
                    unique = True
                else:
                    unique = False

        elif param == "state":
            data = data[0]
            tem_dict = {}
            for k in range(len(data)):
                pr_dict, W_dict = {}, {}
                state_data = data[k]
                prior_data = state_data[0][0][0]
                state_prior_par = ["sigma", "alpha", "Mean"]
                W_data = state_data[1][0][0]
                state_W_par = ["Mu_W", "S_W", "iS_W"]

                for l, sub_par in enumerate(state_prior_par):
                    pr_dict[sub_par] = prior_data[l]
                for l, sub_par in enumerate(state_W_par):
                    W_dict[sub_par] = W_data[l]
                if not unique:
                    Omega_data = state_data[2][0][0]
                    state_Omega_par = [
                        "Gam_rate",
                        "Gam_shape",
                    ]
                    Omega_dict = dict()
                    for l, sub_par in enumerate(state_Omega_par):
                        Omega_dict[sub_par] = Omega_data[l]
                    tem_dict[f"state{k+1}"] = {
                        "prior": pr_dict,
                        "W": W_dict,
                        "Omega": Omega_dict,
                    }
                    continue
                tem_dict[f"state{k+1}"] = {
                    "prior": pr_dict,
                    "W": W_dict,
                }
            hmm_dict[param] = tem_dict
        else:
            hmm_dict[param] = np.squeeze(data)

    return hmm_dict


# Calculate Vpath features
def vpath_fo(vpath, K_state):
    state_fo = dict()
    length = len(vpath)
    for state in range(1, K_state + 1):
        state_fo[f"state{state}_fo"] = [np.count_nonzero(vpath == state) / length]

    return pd.DataFrame.from_dict(state_fo)


def vpath_visit(vpath, K_state):
    dic = dict()
    length = len(vpath)
    for state in range(1, K_state + 1):
        vpath_bool = vpath == state
        visit = 0
        continuous = False
        for i in range(length):
            if not continuous:
                if vpath_bool[i] == True:
                    visit += 1
                else:
                    continue
            try:
                if vpath_bool[i + 1] == True:
                    continuous = True
                else:
                    continuous = False
            except IndexError:
                pass
        dic[f"state{state}_visits"] = [visit]

    return pd.DataFrame.from_dict(dic)


def vpath_switch(vpath, K_state):
    dic = dict()
    length = len(vpath)
    change = 0
    for i in range(length - 1):
        if vpath[i] != vpath[i + 1]:
            change += 1
    dic["switch_rate"] = [change / length]

    return pd.DataFrame.from_dict(dic)


def vpath_lifetime(vpath, K_state):
    dic = dict()
    length = len(vpath)
    for state in range(1, K_state + 1):
        lifes = []
        vpath_bool = vpath == state
        life = 0
        for i in range(length):
            if vpath_bool[i] == True:
                life += 1
            elif (vpath_bool[i] == False) & (life > 0):
                lifes.append(life)
                life = 0
            else:
                pass
        dic[f"state{state}_lifetime"] = [lifes]

    return pd.DataFrame.from_dict(dic)


def vpath_interval(vpath, K_state):
    dic = dict()
    length = len(vpath)
    for state in range(1, K_state + 1):
        intervals = []
        vpath_bool = vpath == state
        interval = 0
        for i in range(length):
            if vpath_bool[i] == False:
                interval += 1
            elif (vpath_bool[i] == True) & (interval > 0):
                intervals.append(interval)
                interval = 0
            else:
                pass
        dic[f"state{state}_interval"] = [intervals]

    return pd.DataFrame.from_dict(dic)


def get_mean_chronnectome(df):
    """
    df should contains the state lifetime or interval for each subject
    """
    result = pd.DataFrame(np.zeros_like(df), columns=df.columns, index=df.index)
    for subj in range(len(df.index)):
        for state in range(len(df.columns)):
            vector = np.array(df.iloc[subj, state])
            if vector.size:
                mean_value = vector.mean()
            else:
                mean_value = 0
            result.iloc[subj, state] = mean_value
    return result.astype(float)


def parse_chronnectome(vpath, subj_num, K_state, timepoints):
    """
    Parse chronnectome data and compute various visitation metrics for each subject.

    Parameters
    ----------
    vpath : list or np.ndarray
        A list or NumPy Array containing the visitation paths for all subjects.
    subj_num : int
        The number of subjects/runs in the dataset.
    K_state : int
        The number of states in the HMM model.
    timepoints : int
        The number of timepoints for each subject/run.

    Returns
    -------
    results : dict
        A dictionary containing DataFrames for each visitation metric, with keys corresponding to the metric names
        The metrics include:
        - vpath_fo: Fractional Occupancy
        - vpath_visit: Visitation counts
        - vpath_lifetime: State lifetimes
        - vpath_interval: State intervals
        - vpath_switch: Switching rate

    """
    funcs = [vpath_fo, vpath_visit, vpath_switch, vpath_lifetime, vpath_interval]
    results = {}
    for func in funcs:
        dfs = []
        for i in range(subj_num):
            subj_vpath = vpath[timepoints * i : timepoints * (i + 1)]
            subj_df = func(subj_vpath, K_state)
            subj_df.index = i
            dfs.append(subj_df)
        if func.__name__ in ["vpath_lifetime", "vpath_interval"]:
            results[func.__name__] = get_mean_chronnectome(pd.concat(dfs))
        else:
            results[func.__name__] = pd.concat(dfs)

    return results



import numpy as np
import pandas as pd
from sklearn.cross_decomposition import CCA
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.stats import pearsonr
from neuroimager.plotting import density_scatter


def permutation_test_cca(X, Y, n_components=1, n_permutations=1000, random_state=None):
    """
    Perform a permutation test for Canonical Correlation Analysis (CCA) to assess the significance
    of the correlation between two sets of features.

    Parameters
    ----------
    X : 2D numpy array, shape (n_samples, n_features_X)
        First set of features.
    Y : 2D numpy array, shape (n_samples, n_features_Y)
        Second set of features.
    n_components : int, optional, default: 1
        Number of CCA components to use.
    n_permutations : int, optional, default: 1000
        Number of permutations to perform.
    random_state : int, RandomState instance or None, optional, default: None
        Seed for random number generator.

    Returns
    -------
    cca : CCA object
        Fitted CCA model.
    X_c : Transformed X data from CCA.
    Y_c : Transformed Y data from CCA.
    original_correlations : list of float
        Correlation values between the two linear combinations of features.
    p_values : list of float
        List of p-values for each CCA component.
    """
    np.random.seed(random_state)

    # Fit the CCA model
    cca = CCA(n_components=n_components)
    X_c, Y_c = cca.fit_transform(X, Y)

    # Compute the original correlations
    original_correlations = [
        pearsonr(X_c[:, i], Y_c[:, i])[0] for i in range(n_components)
    ]

    # Perform the permutation test
    count_better_correlations = [0] * n_components
    seeds = np.random.randint(0, 100000, n_permutations)
    rs = []
    p_values = []
    for i, seed in enumerate(seeds):
        np.random.seed(seed)

        ids = np.random.permutation(X.shape[0])
        x_perm = X[ids, :]
        cca_perm = CCA(n_components=n_components)
        X_perm, Y_perm = cca_perm.fit_transform(x_perm, Y)
        rs.append(pd.DataFrame(X_perm).corrwith(pd.DataFrame(Y_perm)))
    r_permuted_values = pd.DataFrame(rs)
    for i in range(n_components):
        p_value = (
            abs(r_permuted_values.iloc[:, i]) > abs(original_correlations[i])
        ).sum() / r_permuted_values.shape[0]
        p_values.append(p_value)

    return cca, X_c, Y_c, original_correlations, p_values


def plot_cca_scatter(transformed_X, transformed_Y, n_components=1, **kwargs):
    """
    Plots Canonical Correlation Analysis (CCA) scatter plot.

    Parameters:
    transformed_X (array-like): Transformed X data from CCA.
    transformed_Y (array-like): Transformed Y data from CCA.
    n_components (int): Number of components to plot.

    Returns:
    None
    """
    fig_size = kwargs.get("fig_size", (10, 10))
    marker = kwargs.get("marker", "o")
    color = kwargs.get("color", "b")
    xlabel = kwargs.get("xlabel", None)
    ylabel = kwargs.get("ylabel", None)
    title_prefix = kwargs.get("title_prefix", "Scatter plot of CCA components")
    layout = kwargs.get("layout", "tight")
    p_values = kwargs.get("p_values", None)
    for i in range(n_components):
        # plt.scatter(transformed_X[:, i], transformed_Y[:, i], marker=marker, color=color, label=f'Component {i+1}')
        density_scatter(transformed_X[:, i], transformed_Y[:, i], display=False)
        corr_coef = np.corrcoef(transformed_X[:, i], transformed_Y[:, i])[0, 1]
        if kwargs.get("xlabel", False) & kwargs.get("ylabel", False):
            title = f"{title_prefix}: {xlabel} vs {ylabel} pearson r = {corr_coef}"
            if p_values:
                title += f" p = {p_values[i]}"
            plt.title(title)
        else:
            title = f"Component {i + 1}: pearson r = {corr_coef}"
            if p_values:
                title += f" p = {p_values[i]}"
            plt.title(title)
        if not xlabel:
            xlabel = "X Component {}".format(i + 1)
        if not ylabel:
            ylabel = "Y Component {}".format(i + 1)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.tight_layout()
        plt.show()


def plot_cca_weights(
    weights,
    labels,
    ax=False,
    display_thresh=0.2,
    show=False,
    min_font_size=6,
    scaling_factor=50,
    **kwargs,
):
    xlim = kwargs.get("xlim", (-1, 1))
    xlabel = kwargs.get("xlabel", "")
    title = kwargs.get("title", "")
    if not ax:
        fig, ax = plt.subplots(figsize=(8, 5))
    weights = weights.flatten()
    # Sort weights and labels for X features
    sorted_indices = np.argsort(weights)
    sorted_weights = weights[sorted_indices]
    if labels is not None:
        sorted_labels = [labels[idx] for idx in sorted_indices]

    # Plot X feature weights
    ax.set_ylabel("Weights")
    ax.set_xlim(xlim[0], xlim[1])
    ax.set_ylim(-0.5, len(weights) + 0.5)
    ax.set_xlabel(xlabel)
    ax.set_title(title)

    texts = []
    for i, label in enumerate(sorted_labels):
        if abs(sorted_weights[i]) < display_thresh:
            continue
        font_size = max(min_font_size, abs(sorted_weights[i]) * scaling_factor)
        ax.text(
            0,
            i,
            label + f"  weights {round(sorted_weights[i],2)}",
            ha="center",
            va="center",
            fontsize=font_size,
        )

    # adjust_text(texts,only_move={ 'text':'X'})
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])

    if show:
        plt.show()

    return ax


def plot_paired_cca_weights(
    weights_X,
    weights_Y,
    display_thresh=0.2,
    scaling_factor=30,
    x_labels=None,
    y_labels=None,
    **kwargs,
):
    """
    Plots Canonical Correlation Analysis (CCA) feature weights.

    Parameters:
    weights_X (array-like): Weights for X features.
    weights_Y (array-like): Weights for Y features.
    x_labels (list, optional): Labels for X features. Defaults to None.
    y_labels (list, optional): Labels for Y features. Defaults to None.

    Returns:
    None
    """
    figsize = kwargs.get("fig_size", (10, 5))
    n_components = weights_X.shape[1]

    for i in range(n_components):
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
        plot_cca_weights(
            weights_X[:, i],
            x_labels,
            ax=ax1,
            display_thresh=display_thresh,
            min_font_size=6,
            scaling_factor=scaling_factor,
            xlim=(-1, 1),
            xlabel="X Features",
            title="Component {} X feature weights".format(i + 1),
        )
        plot_cca_weights(
            weights_Y[:, i],
            y_labels,
            ax=ax2,
            display_thresh=display_thresh,
            min_font_size=6,
            scaling_factor=scaling_factor,
            xlabel="Y Features",
            title="Component {} Y feature weights".format(i + 1),
        )
        plt.tight_layout()
        plt.show()
